// balboa
// Copyright (c) 2018, 2019 DCSO GmbH

#ifndef __MOCK_H
#define __MOCK_H

#include <stdbool.h>
#include <stdint.h>
#include <stdlib.h>
#include <time.h>

#include <engine.h>

db_t* blb_mock_open(void);

#endif
